# -*- coding: utf-8 -*-

from __future__ import absolute_import

__version__ = "1.0.1"

# the public api
from homebase.homebase import (
    user_data_dir, user_cache_dir, user_config_dir,
    user_logs_dir, user_state_dir, site_data_dir, site_config_dir, site_data_dir_list, site_config_dir_list
)
